# Pantomime output generator class

import time
import re
import random

from types import ListType
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# intra-package imports
import Errors

SEMISPACE = '; '
BAR = '|'

bcre = re.compile('boundary="?([^"]+)"?')



class Generator:
    """Generates output from a Message object tree.

    This basic generator writes the message to the given file object as plain
    text.

    """
    def __init__(self, outfp):
        """Create the generator for message flattening.

        outfp is the output file-like object for writing the message to.  It
        must have a write() method.
        """
        self._fp = outfp

    def write(self, msg, unixfrom=1):
        """Print the message object tree rooted at msg to the output file
        specified when the Generator instance was created.

        unixfrom is a flag that forces the printing of a Unix From_ delimiter
        before the first object in the message tree.  If the original message
        has no From_ delimiter, a `standard' one is crafted.  Set this to
        false to inhibit the printing of any From_ delimiter.

        Note that for subobjects, no From_ line is printed.
        """
        if unixfrom:
            ufrom = msg.get_unixfrom()
            if not ufrom:
                ufrom = 'From nobody ' + time.ctime(time.time())
            print >> self._fp, ufrom
        self._writemsg(msg, self._fp)

    #
    # Non-public methods
    #

    def _writemsg(self, msg, fp, enclosing_boundaries=None):
        # recursively print subobjects
        if enclosing_boundaries is None:
            enclosing_boundaries = []
            suppress = 0
        else:
            suppress = 1
        if msg.ismultipart():
            # if the container doesn't have a boundary, we'll create one
            boundary = None
            ctype = msg.get('content-type')
            if ctype:
                mo = bcre.search(ctype)
                if mo:
                    boundary = mo.group(1)
            else:
                boundary = self._makeboundary()
                ctype = 'multipart/mixed; boundary="%s"' % boundary
            if boundary is None:
                # we'll craft our own boundary
                boundary = self._makeboundary()
                ctype += '; boundary="%s"' % boundary
        #
        # Write out all the headers, in the original order
        for header, val in msg.items():
            if header.lower() == 'content-type' and msg.ismultipart():
                # Use our content type instead
                print >> fp, 'Content-Type:', ctype
            elif header.lower() == 'mime-version' and suppress:
                continue
            else:
                print >> fp, header + ':', val
        # blank line separating header from body
        print >> fp
        if msg.ismultipart():
            preamble = getattr(msg, 'preamble', None)
            postamble = getattr(msg, 'postamble', None)
            if preamble:
                print >> fp, preamble,
            for payload in msg.get_payload():
                print >> fp, '--' + boundary
                enclosing_boundaries.append('--' + boundary)
                self._writemsg(payload, fp, enclosing_boundaries)
                del enclosing_boundaries[-1]
            print >> fp, '--' + boundary + '--'
            if postamble:
                print >> fp, postamble,
        else:
            payload = msg.get_payload()
            # Escape the boundaries
            if enclosing_boundaries:
                cre = re.compile(
                    '^(' +
                    BAR.join([re.escape(x) for x in enclosing_boundaries]) +
                    ')', re.MULTILINE)
                payload = cre.sub(r'- \1', payload)
            print >> fp, payload,
            
                    
    def _makeboundary(self):
        # Craft a random and unlikely multipart/* boundary string
        return ('=' * 20) + repr(random.random()).split('.')[1] + '==_'



if __name__ == '__main__':
    from Parser import Parser
    p = Parser()
    fp = open('/tmp/msg3.txt')
    m = p.parse(fp)
    fp.close()
    import sys
    g = Generator(sys.stdout)
    g.write(m)
