import re

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# intra-package imports
import Errors
import Message

bcre = re.compile('boundary="?([^"]+)"?', re.IGNORECASE)
EMPTYSTRING = ''
NL = '\n'



class Parser:
    def __init__(self, _class=Message.Message):
        """Parser of RFC822 and MIME email messages.

        Creates an in-memory object tree representing the message, which can
        then be manipulated and turned over to a Generator to return the
        textual representation of the message.

        The string must be formatted as a block of rfc822 headers and header
        continuation lines, optionally preceeded by a `Unix-from' header.  The
        header block is terminated either by the end of the string or by a
        blank line.

        _class is the class to instantiate for new message objects when they
        must be created.  This class must have a constructor that can take
        zero arguments.  Default is Message.Message.
        """
        self._class = _class

    def parse(self, fp):
        root = self._class()
        self._parse(root, 0, fp)
        return root

    def parsestr(self, text):
        root = self._class()
        self._parse(root, 0, StringIO(text))
        return root

    def _parse(self, container, lineno, fp):
        # Parse the headers
        lastheader = ''
        lastvalue = []
        while 1:
            line = fp.readline()[:-1]
            if not line or not line.strip():
                break
            lineno += 1
            # Check for initial Unix From_ line
            if line.startswith('From '):
                if lineno == 1:
                    container.set_unixfrom(line)
                    continue
                else:
                    raise Errors.HeaderParseError(
                        'Unix-from in headers after first rfc822 header',
                        lineno)
            #
            # Header continuation line
            if line[0] in ' \t':
                if not lastheader:
                    raise Error.HeaderParseError(
                        'Continuation line seen before first header',
                        lineno)
                lastvalue.append(line)
                continue
            # Normal, non-continuation header
            i = line.find(':')
            if i < 0:
                raise Errors.HeaderParseError(
                    'Not a header, not a continuation', lineno)
            if lastheader:
                container[lastheader] = NL.join(lastvalue)
            lastheader = line[:i]
            lastvalue = [line[i+1:].lstrip()]
        #
        # make sure we retain the last header
        if lastheader:
            container[lastheader] = NL.join(lastvalue)
        #
        # Now parse the body, but first split the payload on the content-type
        # boundary if present
        boundary = None
        ctype = container['content-type']
        if ctype:
            mo = bcre.search(ctype)
            if mo:
                boundary = mo.group(1)
        #
        # If there's a boundary, split the payload text into its constituent
        # parts and parse each separately.  Otherwise, just parse the rest of
        # the body as a single message.  Note: any exceptions raised in the
        # recursive parse need to have their line numbers coerced.
        if boundary:
            preamble = postamble = None
            # Split into subparts.  TBD: is there a faster way?
            separator = '--' + boundary
            payload = fp.read()
            start = payload.find(separator + '\n')
            if start < 0:
                raise Errors.BoundaryError(
                    "Couldn't find starting boundary", lineno)
            if start > 0:
                # there's some pre-MIME boundary preamble
                preamble = payload[0:start]
            start += len(separator) + 1
            terminator = payload.find(separator + '--\n', start)
            if terminator < 0:
                raise Errors.BoundaryError(
                    "Couldn't find terminating boundary",
                    lineno+payload.count('\n'))
            if terminator+len(separator)+3 < len(payload):
                # there's some post-MIME boundary postamble
                postamble = payload[terminator+len(separator)+3:]
            parts = payload[start:terminator].split(separator + '\n')
            for part in parts:
                try:
                    msgobj = self.parsestr(part)
                except Errors.MessageError, e:
                    e += lineno
                    raise
                container.preamble = preamble
                container.postamble = postamble
                container.add_payload(msgobj)
        else:
            container.add_payload(fp.read())



if __name__ == '__main__':
    p = Parser()
    fp = open('/tmp/msg3.txt')
    m = p.parse(fp)
    fp.close()
